@@ -16,8 +16,9 @@ use rayon::prelude::*;
 use crate::{
     graph_ops::{DeserializeGraphOp, SerializeGraphOp},
     index::{AtomicIdx, Idx},
-    input::{edgelist::EdgeList, Direction, DotGraph},
-    DirectedGraph, Error, Graph, SharedMut, UndirectedGraph,
+    input::{edgelist::EdgeList, Direction, DotGraph, MyCypherValue},
+    DirectedDegrees, DirectedNeighbors, DirectedNeighborsWithValues, Error, Graph, SharedMut,
+    UndirectedDegrees, UndirectedNeighbors, UndirectedNeighborsWithValues,
 };
 
 /// Defines how the neighbor list of individual nodes are organized within the
@@ -51,13 +52,47 @@ impl Default for CsrLayout {
 /// neighbor list is defined by `offsets[u + 1] - offsets[u]`. The neighbor list
 /// of `u` is defined by the slice `&targets[offsets[u]..offsets[u + 1]]`.
 #[derive(Debug)]
-pub struct Csr<Index: Idx, T = Index> {
+pub struct Csr<Index: Idx, NI, EV> {
     offsets: Box<[Index]>,
-    targets: Box<[T]>,
+    targets: Box<[Target<NI, EV>]>,
 }
 
-impl<Index: Idx, T> Csr<Index, T> {
-    pub(crate) fn new(offsets: Box<[Index]>, targets: Box<[T]>) -> Self {
+/// Represents the target of an edge and its associated value.
+#[derive(Clone, Copy, Debug)]
+#[repr(C)]
+pub struct Target<NI, EV> {
+    pub target: NI,
+    pub value: EV,
+}
+
+impl<T: Ord, V> Ord for Target<T, V> {
+    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
+        self.target.cmp(&other.target)
+    }
+}
+
+impl<T: PartialOrd, V> PartialOrd for Target<T, V> {
+    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
+        self.target.partial_cmp(&other.target)
+    }
+}
+
+impl<T: PartialEq, V> PartialEq for Target<T, V> {
+    fn eq(&self, other: &Self) -> bool {
+        self.target.eq(&other.target)
+    }
+}
+
+impl<T: Eq, V> Eq for Target<T, V> {}
+
+impl<T, EV> Target<T, EV> {
+    pub fn new(target: T, value: EV) -> Self {
+        Self { target, value }
+    }
+}
+
+impl<Index: Idx, NI, EV> Csr<Index, NI, EV> {
+    pub(crate) fn new(offsets: Box<[Index]>, targets: Box<[Target<NI, EV>]>) -> Self {
         Self { offsets, targets }
     }
 
@@ -80,7 +115,7 @@ impl<Index: Idx, T> Csr<Index, T> {
     }
 
     #[inline]
-    pub(crate) fn targets(&self, i: Index) -> &[T] {
+    pub(crate) fn targets_with_values(&self, i: Index) -> &[Target<NI, EV>] {
         let from = self.offsets[i.index()];
         let to = self.offsets[(i + Index::new(1)).index()];
 
@@ -88,10 +123,39 @@ impl<Index: Idx, T> Csr<Index, T> {
     }
 }
 
-type CsrInput<'a, Node> = (&'a mut EdgeList<Node>, Node, Direction, CsrLayout);
+impl<Index: Idx, NI> Csr<Index, NI, ()> {
+    #[inline]
+    pub(crate) fn targets(&self, i: Index) -> &[NI] {
+        assert_eq!(
+            std::mem::size_of::<Target<NI, ()>>(),
+            std::mem::size_of::<NI>()
+        );
+        assert_eq!(
+            std::mem::align_of::<Target<NI, ()>>(),
+            std::mem::align_of::<NI>()
+        );
+        let from = self.offsets[i.index()];
+        let to = self.offsets[(i + Index::new(1)).index()];
 
-impl<Node: Idx> From<CsrInput<'_, Node>> for Csr<Node> {
-    fn from((edge_list, node_count, direction, csr_layout): CsrInput<'_, Node>) -> Self {
+        let len = (to - from).index();
+
+        let targets = &self.targets[from.index()..to.index()];
+
+        // SAFETY: len is within bounds as it is calculated above as `to - from`.
+        //         The types Target<T, ()> and T are verified to have the same
+        //         size and alignment.
+        unsafe { std::slice::from_raw_parts(targets.as_ptr() as *const _, len) }
+    }
+}
+
+type CsrInput<'a, NI, EV> = (&'a mut EdgeList<NI, EV>, NI, Direction, CsrLayout);
+
+impl<NI, EV> From<CsrInput<'_, NI, EV>> for Csr<NI, NI, EV>
+where
+    NI: Idx,
+    EV: Copy + Send + Sync,
+{
+    fn from((edge_list, node_count, direction, csr_layout): CsrInput<'_, NI, EV>) -> Self {
         let start = Instant::now();
         let degrees = edge_list.degrees(node_count, direction);
         info!("Computed degrees in {:?}", start.elapsed());
@@ -102,7 +166,7 @@ impl<Node: Idx> From<CsrInput<'_, Node>> for Csr<Node> {
 
         let start = Instant::now();
         let edge_count = offsets[node_count.index()].load(Acquire).index();
-        let mut targets = Vec::<Node>::with_capacity(edge_count);
+        let mut targets = Vec::<Target<NI, EV>>::with_capacity(edge_count);
         let targets_ptr = SharedMut::new(targets.as_mut_ptr());
 
         // The following loop writes all targets into their correct position.
@@ -114,20 +178,20 @@ impl<Node: Idx> From<CsrInput<'_, Node>> for Csr<Node> {
         // only write once into each position and every thread that might run
         // will write into different positions.
         if matches!(direction, Direction::Outgoing | Direction::Undirected) {
-            edge_list.par_iter().for_each(|(s, t)| {
+            edge_list.par_iter().for_each(|(s, t, v)| {
                 let offset = offsets[s.index()].get_and_increment(Acquire);
 
                 unsafe {
-                    targets_ptr.add(offset.index()).write(*t);
+                    targets_ptr.add(offset.index()).write(Target::new(*t, *v));
                 }
             })
         }
 
         if matches!(direction, Direction::Incoming | Direction::Undirected) {
-            edge_list.par_iter().for_each(|(s, t)| {
+            edge_list.par_iter().for_each(|(s, t, v)| {
                 let offset = offsets[t.index()].get_and_increment(Acquire);
                 unsafe {
-                    targets_ptr.add(offset.index()).write(*s);
+                    targets_ptr.add(offset.index()).write(Target::new(*s, *v));
                 }
             })
         }
@@ -141,15 +205,15 @@ impl<Node: Idx> From<CsrInput<'_, Node>> for Csr<Node> {
         info!("Computed target array in {:?}", start.elapsed());
 
         let start = Instant::now();
-        // SAFETY: Node and Node::Atomic have the same memory layout
-        let mut offsets = unsafe { transmute::<_, Vec<Node>>(offsets) };
+        // SAFETY: NI and NI::Atomic have the same memory layout
+        let mut offsets = unsafe { transmute::<_, Vec<NI>>(offsets) };
 
         // Each insert into the target array in the previous loops incremented
         // the offset for the corresponding node by one. As a consequence the
         // offset values are shifted one index to the right. We need to correct
         // this in order to get correct offsets.
         offsets.rotate_right(1);
-        offsets[0] = Node::zero();
+        offsets[0] = NI::zero();
         info!("Finalized offset array in {:?}", start.elapsed());
 
         let (offsets, targets) = match csr_layout {
@@ -162,7 +226,7 @@ impl<Node: Idx> From<CsrInput<'_, Node>> for Csr<Node> {
             }
             CsrLayout::Deduplicated => {
                 let start = Instant::now();
-                let offsets_targets = sort_and_deduplicate_targets(&offsets, &mut targets);
+                let offsets_targets = sort_and_deduplicate_targets(&offsets, &mut targets[..]);
                 info!("Sorted and deduplicated targets in {:?}", start.elapsed());
                 offsets_targets
             }
@@ -175,9 +239,37 @@ impl<Node: Idx> From<CsrInput<'_, Node>> for Csr<Node> {
     }
 }
 
-impl<Node: Idx + ToByteSlice> Csr<Node> {
+unsafe impl<NI, EV> ToByteSlice for Target<NI, EV>
+where
+    NI: ToByteSlice,
+    EV: ToByteSlice,
+{
+    fn to_byte_slice<S: AsRef<[Self]> + ?Sized>(slice: &S) -> &[u8] {
+        let slice = slice.as_ref();
+        let len = slice.len() * std::mem::size_of::<Target<NI, EV>>();
+        unsafe { std::slice::from_raw_parts(slice.as_ptr() as *const u8, len) }
+    }
+}
+
+unsafe impl<NI, EV> ToMutByteSlice for Target<NI, EV>
+where
+    NI: ToMutByteSlice,
+    EV: ToMutByteSlice,
+{
+    fn to_mut_byte_slice<S: AsMut<[Self]> + ?Sized>(slice: &mut S) -> &mut [u8] {
+        let slice = slice.as_mut();
+        let len = slice.len() * std::mem::size_of::<Target<NI, EV>>();
+        unsafe { std::slice::from_raw_parts_mut(slice.as_mut_ptr() as *mut u8, len) }
+    }
+}
+
+impl<NI, EV> Csr<NI, NI, EV>
+where
+    NI: Idx + ToByteSlice,
+    EV: ToByteSlice,
+{
     fn serialize<W: Write>(&self, output: &mut W) -> Result<(), Error> {
-        let type_name = std::any::type_name::<Node>().as_bytes();
+        let type_name = std::any::type_name::<NI>().as_bytes();
         output.write_all([type_name.len()].as_byte_slice())?;
         output.write_all(type_name)?;
 
@@ -193,8 +285,12 @@ impl<Node: Idx + ToByteSlice> Csr<Node> {
     }
 }
 
-impl<Node: Idx + ToMutByteSlice> Csr<Node> {
-    fn deserialize<R: Read>(read: &mut R) -> Result<Csr<Node>, Error> {
+impl<NI, EV> Csr<NI, NI, EV>
+where
+    NI: Idx + ToMutByteSlice,
+    EV: ToMutByteSlice,
+{
+    fn deserialize<R: Read>(read: &mut R) -> Result<Csr<NI, NI, EV>, Error> {
         let mut type_name_len = [0_usize; 1];
         read.read_exact(type_name_len.as_mut_byte_slice())?;
         let [type_name_len] = type_name_len;
@@ -203,7 +299,7 @@ impl<Node: Idx + ToMutByteSlice> Csr<Node> {
         read.read_exact(type_name.as_mut_byte_slice())?;
         let type_name = String::from_utf8(type_name).expect("could not read type name");
 
-        let expected_type_name = std::any::type_name::<Node>().to_string();
+        let expected_type_name = std::any::type_name::<NI>().to_string();
 
         if type_name != expected_type_name {
             return Err(Error::InvalidIdType {
@@ -212,19 +308,19 @@ impl<Node: Idx + ToMutByteSlice> Csr<Node> {
             });
         }
 
-        let mut meta = [Node::zero(); 2];
+        let mut meta = [NI::zero(); 2];
         read.read_exact(meta.as_mut_byte_slice())?;
 
         let [node_count, edge_count] = meta;
 
-        let mut offsets = Box::<[Node]>::new_uninit_slice(node_count.index() + 1);
-        let offsets_ptr = offsets.as_mut_ptr() as *mut Node;
+        let mut offsets = Box::<[NI]>::new_uninit_slice(node_count.index() + 1);
+        let offsets_ptr = offsets.as_mut_ptr() as *mut NI;
         let offsets_ptr =
             unsafe { std::slice::from_raw_parts_mut(offsets_ptr, node_count.index() + 1) };
         read.read_exact(offsets_ptr.as_mut_byte_slice())?;
 
-        let mut targets = Box::<[Node]>::new_uninit_slice(edge_count.index());
-        let targets_ptr = targets.as_mut_ptr() as *mut Node;
+        let mut targets = Box::<[Target<NI, EV>]>::new_uninit_slice(edge_count.index());
+        let targets_ptr = targets.as_mut_ptr() as *mut Target<NI, EV>;
         let targets_ptr =
             unsafe { std::slice::from_raw_parts_mut(targets_ptr, edge_count.index()) };
         read.read_exact(targets_ptr.as_mut_byte_slice())?;
@@ -236,64 +332,71 @@ impl<Node: Idx + ToMutByteSlice> Csr<Node> {
     }
 }
 
-pub struct DirectedCsrGraph<Node: Idx> {
-    node_count: Node,
-    edge_count: Node,
-    csr_out: Csr<Node>,
-    csr_inc: Csr<Node>,
+pub struct DirectedCsrGraph<NI: Idx, EV = ()> {
+    csr_out: Csr<NI, NI, EV>,
+    csr_inc: Csr<NI, NI, EV>,
 }
 
-impl<Node: Idx> DirectedCsrGraph<Node> {
-    pub fn new(csr_out: Csr<Node>, csr_inc: Csr<Node>) -> Self {
-        let node_count = csr_out.node_count();
-        let edge_count = csr_out.edge_count();
-
+impl<NI: Idx, EV> DirectedCsrGraph<NI, EV> {
+    pub fn new(csr_out: Csr<NI, NI, EV>, csr_inc: Csr<NI, NI, EV>) -> Self {
+        let g = Self { csr_out, csr_inc };
         info!(
             "Created directed graph (node_count = {:?}, edge_count = {:?})",
-            node_count, edge_count
+            g.node_count(),
+            g.edge_count()
         );
 
-        Self {
-            node_count,
-            edge_count,
-            csr_out,
-            csr_inc,
+        g
+    }
+}
+
+impl<NI: Idx, EV> Graph<NI> for DirectedCsrGraph<NI, EV> {
+    delegate::delegate! {
+        to self.csr_out {
+            fn node_count(&self) -> NI;
+            fn edge_count(&self) -> NI;
         }
     }
 }
 
-impl<Node: Idx> Graph<Node> for DirectedCsrGraph<Node> {
-    fn node_count(&self) -> Node {
-        self.node_count
-    }
-
-    fn edge_count(&self) -> Node {
-        self.edge_count
-    }
-}
-
-impl<Node: Idx> DirectedGraph<Node> for DirectedCsrGraph<Node> {
-    fn out_degree(&self, node: Node) -> Node {
+impl<NI: Idx, EV> DirectedDegrees<NI> for DirectedCsrGraph<NI, EV> {
+    fn out_degree(&self, node: NI) -> NI {
         self.csr_out.degree(node)
     }
 
-    fn out_neighbors(&self, node: Node) -> &[Node] {
+    fn in_degree(&self, node: NI) -> NI {
+        self.csr_inc.degree(node)
+    }
+}
+
+impl<NI: Idx> DirectedNeighbors<NI> for DirectedCsrGraph<NI, ()> {
+    fn out_neighbors(&self, node: NI) -> &[NI] {
         self.csr_out.targets(node)
     }
 
-    fn in_degree(&self, node: Node) -> Node {
-        self.csr_inc.degree(node)
-    }
-
-    fn in_neighbors(&self, node: Node) -> &[Node] {
+    fn in_neighbors(&self, node: NI) -> &[NI] {
         self.csr_inc.targets(node)
     }
 }
 
-impl<Node: Idx> From<(EdgeList<Node>, CsrLayout)> for DirectedCsrGraph<Node> {
-    fn from((mut edge_list, csr_option): (EdgeList<Node>, CsrLayout)) -> Self {
+impl<NI: Idx, EV> DirectedNeighborsWithValues<NI, EV> for DirectedCsrGraph<NI, EV> {
+    fn out_neighbors_with_values(&self, node: NI) -> &[Target<NI, EV>] {
+        self.csr_out.targets_with_values(node)
+    }
+
+    fn in_neighbors_with_values(&self, node: NI) -> &[Target<NI, EV>] {
+        self.csr_inc.targets_with_values(node)
+    }
+}
+
+impl<NI, EV> From<(EdgeList<NI, EV>, CsrLayout)> for DirectedCsrGraph<NI, EV>
+where
+    NI: Idx,
+    EV: Copy + Send + Sync,
+{
+    fn from((mut edge_list, csr_option): (EdgeList<NI, EV>, CsrLayout)) -> Self {
         info!("Creating directed graph");
-        let node_count = edge_list.max_node_id() + Node::new(1);
+        let node_count = edge_list.max_node_id() + NI::new(1);
 
         let start = Instant::now();
         let csr_out = Csr::from((&mut edge_list, node_count, Direction::Outgoing, csr_option));
@@ -307,12 +410,12 @@ impl<Node: Idx> From<(EdgeList<Node>, CsrLayout)> for DirectedCsrGraph<Node> {
     }
 }
 
-impl<Node, Label> From<(DotGraph<Node, Label>, CsrLayout)> for DirectedCsrGraph<Node>
+impl<NI, Label> From<(DotGraph<NI, Label>, CsrLayout)> for DirectedCsrGraph<NI, ()>
 where
-    Node: Idx,
+    NI: Idx,
     Label: Idx + Hash,
 {
-    fn from((dot_graph, csr_layout): (DotGraph<Node, Label>, CsrLayout)) -> Self {
+    fn from((dot_graph, csr_layout): (DotGraph<NI, Label>, CsrLayout)) -> Self {
         let DotGraph {
             label_frequencies: _,
             edge_list,
@@ -325,26 +428,34 @@ where
     }
 }
 
-impl<Node: Idx> From<(&gdl::Graph, CsrLayout)> for DirectedCsrGraph<Node> {
-    fn from((gdl_graph, csr_layout): (&gdl::Graph, CsrLayout)) -> Self {
+impl<'a, NI, EV> From<(&'a gdl::Graph, CsrLayout)> for DirectedCsrGraph<NI, EV>
+where
+    NI: Idx,
+    EV: From<MyCypherValue<'a>> + Default + Copy + Send + Sync,
+{
+    fn from((gdl_graph, csr_layout): (&'a gdl::Graph, CsrLayout)) -> Self {
         DirectedCsrGraph::from((EdgeList::from(gdl_graph), csr_layout))
     }
 }
 
-impl<Node: Idx> From<(gdl::Graph, CsrLayout)> for DirectedCsrGraph<Node> {
+impl<NI, EV> From<(gdl::Graph, CsrLayout)> for DirectedCsrGraph<NI, EV>
+where
+    NI: Idx,
+    for<'a> EV: From<MyCypherValue<'a>> + Default + Copy + Send + Sync,
+{
     fn from((gdl_graph, csr_layout): (gdl::Graph, CsrLayout)) -> Self {
         DirectedCsrGraph::from((EdgeList::from(&gdl_graph), csr_layout))
     }
 }
 
-impl<W: Write, Node: Idx + ToByteSlice> SerializeGraphOp<W> for DirectedCsrGraph<Node> {
+impl<W, NI, EV> SerializeGraphOp<W> for DirectedCsrGraph<NI, EV>
+where
+    W: Write,
+    NI: Idx + ToByteSlice,
+    EV: ToByteSlice,
+{
     fn serialize(&self, mut output: W) -> Result<(), Error> {
-        let DirectedCsrGraph {
-            node_count: _,
-            edge_count: _,
-            csr_out,
-            csr_inc,
-        } = self;
+        let DirectedCsrGraph { csr_out, csr_inc } = self;
 
         csr_out.serialize(&mut output)?;
         csr_inc.serialize(&mut output)?;
@@ -353,15 +464,24 @@ impl<W: Write, Node: Idx + ToByteSlice> SerializeGraphOp<W> for DirectedCsrGraph
     }
 }
 
-impl<R: Read, Node: Idx + ToMutByteSlice> DeserializeGraphOp<R, Self> for DirectedCsrGraph<Node> {
+impl<R, NI, EV> DeserializeGraphOp<R, Self> for DirectedCsrGraph<NI, EV>
+where
+    R: Read,
+    NI: Idx + ToMutByteSlice,
+    EV: ToMutByteSlice,
+{
     fn deserialize(mut read: R) -> Result<Self, Error> {
-        let csr_out: Csr<Node> = Csr::deserialize(&mut read)?;
-        let csr_inc: Csr<Node> = Csr::deserialize(&mut read)?;
+        let csr_out: Csr<NI, NI, EV> = Csr::deserialize(&mut read)?;
+        let csr_inc: Csr<NI, NI, EV> = Csr::deserialize(&mut read)?;
         Ok(DirectedCsrGraph::new(csr_out, csr_inc))
     }
 }
 
-impl<Node: Idx + ToMutByteSlice> TryFrom<(PathBuf, CsrLayout)> for DirectedCsrGraph<Node> {
+impl<NI, EV> TryFrom<(PathBuf, CsrLayout)> for DirectedCsrGraph<NI, EV>
+where
+    NI: Idx + ToMutByteSlice,
+    EV: ToMutByteSlice,
+{
     type Error = Error;
 
     fn try_from((path, _): (PathBuf, CsrLayout)) -> Result<Self, Self::Error> {
@@ -373,60 +493,65 @@ impl<Node: Idx + ToMutByteSlice> TryFrom<(PathBuf, CsrLayout)> for DirectedCsrGr
 }
 
 #[derive(Debug)]
-pub struct UndirectedCsrGraph<Node: Idx> {
-    node_count: Node,
-    edge_count: Node,
-    csr: Csr<Node>,
+pub struct UndirectedCsrGraph<NI: Idx, EV = ()> {
+    csr: Csr<NI, NI, EV>,
 }
 
-impl<Node: Idx> From<Csr<Node>> for UndirectedCsrGraph<Node> {
-    fn from(csr: Csr<Node>) -> Self {
+impl<NI: Idx, EV> From<Csr<NI, NI, EV>> for UndirectedCsrGraph<NI, EV> {
+    fn from(csr: Csr<NI, NI, EV>) -> Self {
         UndirectedCsrGraph::new(csr)
     }
 }
 
-impl<Node: Idx> UndirectedCsrGraph<Node> {
-    pub fn new(csr: Csr<Node>) -> Self {
-        let node_count = csr.node_count();
-        let edge_count = csr.edge_count() / Node::new(2);
-
+impl<NI: Idx, EV> UndirectedCsrGraph<NI, EV> {
+    pub fn new(csr: Csr<NI, NI, EV>) -> Self {
+        let g = Self { csr };
         info!(
             "Created undirected graph (node_count = {:?}, edge_count = {:?})",
-            node_count, edge_count
+            g.node_count(),
+            g.edge_count()
         );
 
-        Self {
-            node_count,
-            edge_count,
-            csr,
-        }
+        g
     }
 }
 
-impl<Node: Idx> Graph<Node> for UndirectedCsrGraph<Node> {
-    fn node_count(&self) -> Node {
-        self.node_count
+impl<NI: Idx, EV> Graph<NI> for UndirectedCsrGraph<NI, EV> {
+    fn node_count(&self) -> NI {
+        self.csr.node_count()
     }
 
-    fn edge_count(&self) -> Node {
-        self.edge_count
+    fn edge_count(&self) -> NI {
+        self.csr.edge_count() / NI::new(2)
     }
 }
 
-impl<Node: Idx> UndirectedGraph<Node> for UndirectedCsrGraph<Node> {
-    fn degree(&self, node: Node) -> Node {
+impl<NI: Idx, EV> UndirectedDegrees<NI> for UndirectedCsrGraph<NI, EV> {
+    fn degree(&self, node: NI) -> NI {
         self.csr.degree(node)
     }
+}
 
-    fn neighbors(&self, node: Node) -> &[Node] {
+impl<NI: Idx> UndirectedNeighbors<NI> for UndirectedCsrGraph<NI, ()> {
+    fn neighbors(&self, node: NI) -> &[NI] {
         self.csr.targets(node)
     }
 }
 
-impl<Node: Idx> From<(EdgeList<Node>, CsrLayout)> for UndirectedCsrGraph<Node> {
-    fn from((mut edge_list, csr_option): (EdgeList<Node>, CsrLayout)) -> Self {
+impl<NI: Idx, EV> UndirectedNeighborsWithValues<NI, EV> for UndirectedCsrGraph<NI, EV> {
+    fn neighbors_with_values(&self, node: NI) -> &[Target<NI, EV>] {
+        self.csr.targets_with_values(node)
+    }
+}
+
+impl<NI, EV> From<(EdgeList<NI, EV>, CsrLayout)> for UndirectedCsrGraph<NI, EV>
+where
+    NI: Idx,
+    EV: Copy + Send + Sync,
+{
+    fn from((mut edge_list, csr_option): (EdgeList<NI, EV>, CsrLayout)) -> Self {
         info!("Creating undirected graph");
-        let node_count = edge_list.max_node_id() + Node::new(1);
+        let node_count = edge_list.max_node_id() + NI::new(1);
 
         let start = Instant::now();
         let csr = Csr::from((
@@ -441,12 +566,12 @@ impl<Node: Idx> From<(EdgeList<Node>, CsrLayout)> for UndirectedCsrGraph<Node> {
     }
 }
 
-impl<Node, Label> From<(DotGraph<Node, Label>, CsrLayout)> for UndirectedCsrGraph<Node>
+impl<NI, Label> From<(DotGraph<NI, Label>, CsrLayout)> for UndirectedCsrGraph<NI, ()>
 where
-    Node: Idx,
+    NI: Idx,
     Label: Idx + Hash,
 {
-    fn from((dot_graph, csr_layout): (DotGraph<Node, Label>, CsrLayout)) -> Self {
+    fn from((dot_graph, csr_layout): (DotGraph<NI, Label>, CsrLayout)) -> Self {
         let DotGraph {
             label_frequencies: _,
             edge_list,
@@ -459,25 +584,34 @@ where
     }
 }
 
-impl<Node: Idx> From<(&gdl::Graph, CsrLayout)> for UndirectedCsrGraph<Node> {
-    fn from((gdl_graph, csr_layout): (&gdl::Graph, CsrLayout)) -> Self {
+impl<'a, NI, EV> From<(&'a gdl::Graph, CsrLayout)> for UndirectedCsrGraph<NI, EV>
+where
+    NI: Idx,
+    EV: From<MyCypherValue<'a>> + Default + Copy + Send + Sync,
+{
+    fn from((gdl_graph, csr_layout): (&'a gdl::Graph, CsrLayout)) -> Self {
         UndirectedCsrGraph::from((EdgeList::from(gdl_graph), csr_layout))
     }
 }
 
-impl<Node: Idx> From<(gdl::Graph, CsrLayout)> for UndirectedCsrGraph<Node> {
+impl<NI, EV> From<(gdl::Graph, CsrLayout)> for UndirectedCsrGraph<NI, EV>
+where
+    NI: Idx,
+    for<'a> EV: From<MyCypherValue<'a>> + Default + Copy + Send + Sync,
+{
     fn from((gdl_graph, csr_layout): (gdl::Graph, CsrLayout)) -> Self {
         UndirectedCsrGraph::from((EdgeList::from(&gdl_graph), csr_layout))
     }
 }
 
-impl<W: Write, Node: Idx + ToByteSlice> SerializeGraphOp<W> for UndirectedCsrGraph<Node> {
+impl<W, NI, EV> SerializeGraphOp<W> for UndirectedCsrGraph<NI, EV>
+where
+    W: Write,
+    NI: Idx + ToByteSlice,
+    EV: ToByteSlice,
+{
     fn serialize(&self, mut output: W) -> Result<(), Error> {
-        let UndirectedCsrGraph {
-            node_count: _,
-            edge_count: _,
-            csr,
-        } = self;
+        let UndirectedCsrGraph { csr } = self;
 
         csr.serialize(&mut output)?;
 
@@ -485,14 +619,23 @@ impl<W: Write, Node: Idx + ToByteSlice> SerializeGraphOp<W> for UndirectedCsrGra
     }
 }
 
-impl<R: Read, Node: Idx + ToMutByteSlice> DeserializeGraphOp<R, Self> for UndirectedCsrGraph<Node> {
+impl<R, NI, EV> DeserializeGraphOp<R, Self> for UndirectedCsrGraph<NI, EV>
+where
+    R: Read,
+    NI: Idx + ToMutByteSlice,
+    EV: ToMutByteSlice,
+{
     fn deserialize(mut read: R) -> Result<Self, Error> {
-        let csr: Csr<Node> = Csr::deserialize(&mut read)?;
+        let csr: Csr<NI, NI, EV> = Csr::deserialize(&mut read)?;
         Ok(UndirectedCsrGraph::new(csr))
     }
 }
 
-impl<Node: Idx + ToMutByteSlice> TryFrom<(PathBuf, CsrLayout)> for UndirectedCsrGraph<Node> {
+impl<NI, EV> TryFrom<(PathBuf, CsrLayout)> for UndirectedCsrGraph<NI, EV>
+where
+    NI: Idx + ToMutByteSlice,
+    EV: ToMutByteSlice,
+{
     type Error = Error;
 
     fn try_from((path, _): (PathBuf, CsrLayout)) -> Result<Self, Self::Error> {
@@ -500,11 +643,11 @@ impl<Node: Idx + ToMutByteSlice> TryFrom<(PathBuf, CsrLayout)> for UndirectedCsr
     }
 }
 
-fn prefix_sum_atomic<Node: AtomicIdx>(degrees: Vec<Node>) -> Vec<Node> {
+fn prefix_sum_atomic<NI: AtomicIdx>(degrees: Vec<NI>) -> Vec<NI> {
     let mut last = degrees.last().unwrap().load(Acquire);
     let mut sums = degrees
         .into_iter()
-        .scan(Node::Inner::zero(), |total, degree| {
+        .scan(NI::Inner::zero(), |total, degree| {
             let value = *total;
             *total += degree.into_inner();
             Some(value.atomic())
@@ -517,11 +660,11 @@ fn prefix_sum_atomic<Node: AtomicIdx>(degrees: Vec<Node>) -> Vec<Node> {
     sums
 }
 
-pub(crate) fn prefix_sum<Node: Idx>(degrees: Vec<Node>) -> Vec<Node> {
+pub(crate) fn prefix_sum<NI: Idx>(degrees: Vec<NI>) -> Vec<NI> {
     let mut last = *degrees.last().unwrap();
     let mut sums = degrees
         .into_iter()
-        .scan(Node::zero(), |total, degree| {
+        .scan(NI::zero(), |total, degree| {
             let value = *total;
             *total += degree;
             Some(value)
@@ -532,16 +675,25 @@ pub(crate) fn prefix_sum<Node: Idx>(degrees: Vec<Node>) -> Vec<Node> {
     sums
 }
 
-pub(crate) fn sort_targets<Node: Idx, T: Send + Ord>(offsets: &[Node], targets: &mut [T]) {
+pub(crate) fn sort_targets<NI, T, EV>(offsets: &[NI], targets: &mut [Target<T, EV>])
+where
+    NI: Idx,
+    T: Copy + Send + Ord,
+    EV: Send,
+{
     to_mut_slices(offsets, targets)
         .par_iter_mut()
         .for_each(|list| list.sort_unstable());
 }
 
-fn sort_and_deduplicate_targets<Node: Idx>(
-    offsets: &[Node],
-    targets: &mut [Node],
-) -> (Vec<Node>, Vec<Node>) {
+fn sort_and_deduplicate_targets<NI, EV>(
+    offsets: &[NI],
+    targets: &mut [Target<NI, EV>],
+) -> (Vec<NI>, Vec<Target<NI, EV>>)
+where
+    NI: Idx,
+    EV: Copy + Send,
+{
     let node_count = offsets.len() - 1;
 
     let mut new_degrees = Vec::with_capacity(node_count);
@@ -556,11 +708,11 @@ fn sort_and_deduplicate_targets<Node: Idx>(
             let (dedup, _) = slice.partition_dedup();
             let mut new_degree = dedup.len();
             // remove self loops .. there is at most once occurence of node inside dedup
-            if let Ok(idx) = dedup.binary_search(&Node::new(node)) {
+            if let Ok(idx) = dedup.binary_search_by_key(&NI::new(node), |t| t.target) {
                 dedup[idx..].rotate_left(1);
                 new_degree -= 1;
             }
-            Node::new(new_degree)
+            NI::new(new_degree)
         })
         .collect_into_vec(&mut new_degrees);
 
@@ -568,7 +720,7 @@ fn sort_and_deduplicate_targets<Node: Idx>(
     debug_assert_eq!(new_offsets.len(), node_count + 1);
 
     let edge_count = new_offsets[node_count].index();
-    let mut new_targets: Vec<Node> = Vec::with_capacity(edge_count);
+    let mut new_targets: Vec<Target<NI, EV>> = Vec::with_capacity(edge_count);
     let new_target_slices = to_mut_slices(&new_offsets, new_targets.spare_capacity_mut());
 
     target_slices
@@ -587,8 +739,8 @@ fn sort_and_deduplicate_targets<Node: Idx>(
     (new_offsets, new_targets)
 }
 
-fn to_mut_slices<'targets, Node: Idx, T>(
-    offsets: &[Node],
+fn to_mut_slices<'targets, NI: Idx, T>(
+    offsets: &[NI],
     targets: &'targets mut [T],
 ) -> Vec<&'targets mut [T]> {
     let node_count = offsets.len() - 1;
@@ -629,13 +781,20 @@ mod tests {
         );
     }
 
+    fn t<T>(t: T) -> Target<T, ()> {
+        Target::new(t, ())
+    }
+
     #[test]
     fn sort_targets_test() {
         let offsets = &[0, 2, 5, 5, 8];
-        let mut targets = vec![1, 0, 4, 2, 3, 5, 6, 7];
-        sort_targets::<usize, _>(offsets, &mut targets);
+        let mut targets = vec![t(1), t(0), t(4), t(2), t(3), t(5), t(6), t(7)];
+        sort_targets::<usize, _, _>(offsets, &mut targets);
 
-        assert_eq!(targets, vec![0, 1, 2, 3, 4, 5, 6, 7]);
+        assert_eq!(
+            targets,
+            vec![t(0), t(1), t(2), t(3), t(4), t(5), t(6), t(7)]
+        );
     }
 
     #[test]
@@ -643,11 +802,11 @@ mod tests {
         let offsets = &[0, 3, 7, 7, 10];
         // 0: [1, 1, 0]    => [1] (removed duplicate and self loop)
         // 1: [4, 2, 3, 2] => [2, 3, 4] (removed duplicate)
-        let mut targets = vec![1, 1, 0, 4, 2, 3, 2, 5, 6, 7];
-        let (offsets, targets) = sort_and_deduplicate_targets::<usize>(offsets, &mut targets);
+        let mut targets = vec![t(1), t(1), t(0), t(4), t(2), t(3), t(2), t(5), t(6), t(7)];
+        let (offsets, targets) = sort_and_deduplicate_targets::<usize, _>(offsets, &mut targets);
 
         assert_eq!(offsets, vec![0, 1, 4, 4, 7]);
-        assert_eq!(targets, vec![1, 2, 3, 4, 5, 6, 7]);
+        assert_eq!(targets, vec![t(1), t(2), t(3), t(4), t(5), t(6), t(7)]);
     }
 
     #[test]
