@@ -1,16 +1,19 @@
 use log::info;
 use rayon::prelude::*;
 
-use crate::graph::csr::{prefix_sum, Csr};
+use crate::graph::csr::{prefix_sum, Csr, Target};
 use crate::index::Idx;
-use crate::{DirectedGraph, Error, Graph, SharedMut, UndirectedGraph};
+use crate::{
+    DirectedDegrees, DirectedNeighborsWithValues, Error, Graph, SharedMut, UndirectedDegrees,
+    UndirectedNeighborsWithValues,
+};
 
 use std::ops::Range;
 use std::sync::Arc;
 use std::time::Instant;
 
 /// Partition the node set based on the degrees of the nodes.
-pub trait DegreePartitionOp<Node: Idx> {
+pub trait DegreePartitionOp<NI: Idx, EV> {
     /// Creates a range-based degree partition of the nodes.
     ///
     /// Divide the nodes into `concurrency` number of ranges such that these
@@ -18,11 +21,11 @@ pub trait DegreePartitionOp<Node: Idx> {
     /// of the nodes of each range should be roughly equal to the extent that
     /// that's actually possible.
     /// The length of the returned vector will never exceed `concurrency`.
-    fn degree_partition(&self, concurrency: usize) -> Vec<Range<Node>>;
+    fn degree_partition(&self, concurrency: usize) -> Vec<Range<NI>>;
 }
 
 /// Partition the node set based on the out degrees of the nodes.
-pub trait OutDegreePartitionOp<Node: Idx> {
+pub trait OutDegreePartitionOp<NI: Idx, EV> {
     /// Creates a range-based out degree partition of the nodes.
     ///
     /// Divide the nodes into `concurrency` number of ranges such that these
@@ -30,11 +33,11 @@ pub trait OutDegreePartitionOp<Node: Idx> {
     /// degrees of the nodes of each range should be roughly equal to the extent
     /// that that's actually possible.
     /// The length of the returned vector will never exceed `concurrency`.
-    fn out_degree_partition(&self, concurrency: usize) -> Vec<Range<Node>>;
+    fn out_degree_partition(&self, concurrency: usize) -> Vec<Range<NI>>;
 }
 
 /// Partition the node set based on the in degrees of the nodes.
-pub trait InDegreePartitionOp<Node: Idx> {
+pub trait InDegreePartitionOp<NI: Idx, EV> {
     /// Creates a range-based in degree partition of the nodes.
     ///
     /// Divide the nodes into `concurrency` number of ranges such that these
@@ -42,11 +45,11 @@ pub trait InDegreePartitionOp<Node: Idx> {
     /// degrees of the nodes of each range should be roughly equal to the extent
     /// that that's actually possible.
     /// The length of the returned vector will never exceed `concurrency`.
-    fn in_degree_partition(&self, concurrency: usize) -> Vec<Range<Node>>;
+    fn in_degree_partition(&self, concurrency: usize) -> Vec<Range<NI>>;
 }
 
 /// Call a particular function for each node with its corresponding state in parallel.
-pub trait ForEachNodeParallelOp<Node: Idx> {
+pub trait ForEachNodeParallelOp<NI: Idx> {
     /// For each node calls `node_fn` with the node and its corresponding mutable
     /// state in parallel.
     ///
@@ -78,12 +81,12 @@ pub trait ForEachNodeParallelOp<Node: Idx> {
     fn for_each_node_par<T, F>(&self, node_values: &mut [T], node_fn: F) -> Result<(), Error>
     where
         T: Send,
-        F: Fn(&Self, Node, &mut T) + Send + Sync;
+        F: Fn(&Self, NI, &mut T) + Send + Sync;
 }
 
 /// Call a particular function for each node with its corresponding state in parallel based on a
 /// partition.
-pub trait ForEachNodeParallelByPartitionOp<Node: Idx> {
+pub trait ForEachNodeParallelByPartitionOp<NI: Idx> {
     /// For each node calls `node_fn` with the node and its corresponding
     /// mutable state in parallel, using `partition` as a parallelization hint.
     ///
@@ -119,16 +122,16 @@ pub trait ForEachNodeParallelByPartitionOp<Node: Idx> {
     /// ```
     fn for_each_node_par_by_partition<T, F>(
         &self,
-        partition: &[Range<Node>],
+        partition: &[Range<NI>],
         node_values: &mut [T],
         node_fn: F,
     ) -> Result<(), Error>
     where
         T: Send,
-        F: Fn(&Self, Node, &mut T) + Send + Sync;
+        F: Fn(&Self, NI, &mut T) + Send + Sync;
 }
 
-pub trait RelabelByDegreeOp<Node: Idx> {
+pub trait RelabelByDegreeOp<N, EV> {
     /// Creates a new graph by relabeling the node ids of the given graph.
     ///
     /// Ids are relabaled using descending degree-order, i.e., given `n` nodes,
@@ -174,20 +177,25 @@ pub trait DeserializeGraphOp<R, G> {
     fn deserialize(read: R) -> Result<G, Error>;
 }
 
-impl<Node, G> RelabelByDegreeOp<Node> for G
+impl<G, NI, EV> RelabelByDegreeOp<NI, EV> for G
 where
-    Node: Idx,
-    G: From<Csr<Node>> + UndirectedGraph<Node> + Sync,
+    NI: Idx,
+    EV: Copy + Ord + Sync,
+    G: Graph<NI>
+        + UndirectedDegrees<NI>
+        + UndirectedNeighborsWithValues<NI, EV>
+        + From<Csr<NI, NI, EV>>
+        + Sync,
 {
     fn to_degree_ordered(&self) -> Self {
         relabel_by_degree(self)
     }
 }
 
-impl<Node, G> ForEachNodeParallelOp<Node> for G
+impl<NI, G> ForEachNodeParallelOp<NI> for G
 where
-    Node: Idx,
-    G: Graph<Node> + Sync,
+    NI: Idx,
+    G: Graph<NI> + Sync,
 {
     /// For each node calls a given function with the node and its corresponding
     /// mutable state in parallel.
@@ -197,7 +205,7 @@ where
     fn for_each_node_par<T, F>(&self, node_values: &mut [T], node_fn: F) -> Result<(), Error>
     where
         T: Send,
-        F: Fn(&Self, Node, &mut T) + Send + Sync,
+        F: Fn(&Self, NI, &mut T) + Send + Sync,
     {
         if node_values.len() != self.node_count().index() {
             return Err(Error::InvalidNodeValues);
@@ -208,16 +216,16 @@ where
         node_values
             .into_par_iter()
             .enumerate()
-            .for_each(|(i, node_state)| node_fn(self, Node::new(i), node_state));
+            .for_each(|(i, node_state)| node_fn(self, NI::new(i), node_state));
 
         Ok(())
     }
 }
 
-impl<Node, G> ForEachNodeParallelByPartitionOp<Node> for G
+impl<NI, G> ForEachNodeParallelByPartitionOp<NI> for G
 where
-    Node: Idx,
-    G: Graph<Node> + Sync,
+    NI: Idx,
+    G: Graph<NI> + Sync,
 {
     /// For each node calls a given function with the node and its corresponding
     /// mutable state in parallel based on the provided node partition.
@@ -226,19 +234,19 @@ where
     /// based fork join with a task for each range in the provided node partition.
     fn for_each_node_par_by_partition<T, F>(
         &self,
-        partition: &[Range<Node>],
+        partition: &[Range<NI>],
         node_values: &mut [T],
         node_fn: F,
     ) -> Result<(), Error>
     where
         T: Send,
-        F: Fn(&Self, Node, &mut T) + Send + Sync,
+        F: Fn(&Self, NI, &mut T) + Send + Sync,
     {
         if node_values.len() != self.node_count().index() {
             return Err(Error::InvalidNodeValues);
         }
 
-        if partition.iter().map(|r| r.end - r.start).sum::<Node>() != self.node_count() {
+        if partition.iter().map(|r| r.end - r.start).sum::<NI>() != self.node_count() {
             return Err(Error::InvalidPartitioning);
         }
 
@@ -259,7 +267,11 @@ where
     }
 }
 
-impl<Node: Idx, U: UndirectedGraph<Node>> DegreePartitionOp<Node> for U {
+impl<NI, EV, U> DegreePartitionOp<NI, EV> for U
+where
+    NI: Idx,
+    U: Graph<NI> + UndirectedDegrees<NI> + UndirectedNeighborsWithValues<NI, EV>,
+{
     /// Creates a greedy range-based degree partition of the nodes.
     ///
     /// It is greedy in the sense that it goes through the node set only once
@@ -281,7 +293,7 @@ impl<Node: Idx, U: UndirectedGraph<Node>> DegreePartitionOp<Node> for U {
     /// assert_eq!(partition[0], 0..1);
     /// assert_eq!(partition[1], 1..4);
     /// ```
-    fn degree_partition(&self, concurrency: usize) -> Vec<Range<Node>> {
+    fn degree_partition(&self, concurrency: usize) -> Vec<Range<NI>> {
         let batch_size = ((self.edge_count().index() * 2) as f64 / concurrency as f64).ceil();
         greedy_node_map_partition(
             |node| self.degree(node).index(),
@@ -292,7 +304,11 @@ impl<Node: Idx, U: UndirectedGraph<Node>> DegreePartitionOp<Node> for U {
     }
 }
 
-impl<Node: Idx, D: DirectedGraph<Node>> OutDegreePartitionOp<Node> for D {
+impl<NI, EV, D> OutDegreePartitionOp<NI, EV> for D
+where
+    NI: Idx,
+    D: Graph<NI> + DirectedDegrees<NI> + DirectedNeighborsWithValues<NI, EV>,
+{
     /// Creates a greedy range-based out degree partition of the nodes.
     ///
     /// It is greedy in the sense that it goes through the node set only once
@@ -314,7 +330,7 @@ impl<Node: Idx, D: DirectedGraph<Node>> OutDegreePartitionOp<Node> for D {
     /// assert_eq!(partition[0], 0..1);
     /// assert_eq!(partition[1], 1..4);
     /// ```
-    fn out_degree_partition(&self, concurrency: usize) -> Vec<Range<Node>> {
+    fn out_degree_partition(&self, concurrency: usize) -> Vec<Range<NI>> {
         let batch_size = (self.edge_count().index() as f64 / concurrency as f64).ceil();
         greedy_node_map_partition(
             |node| self.out_degree(node).index(),
@@ -325,7 +341,11 @@ impl<Node: Idx, D: DirectedGraph<Node>> OutDegreePartitionOp<Node> for D {
     }
 }
 
-impl<Node: Idx, D: DirectedGraph<Node>> InDegreePartitionOp<Node> for D {
+impl<NI, EV, D> InDegreePartitionOp<NI, EV> for D
+where
+    NI: Idx,
+    D: Graph<NI> + DirectedDegrees<NI> + DirectedNeighborsWithValues<NI, EV>,
+{
     /// Creates a greedy range-based in degree partition of the nodes.
     ///
     /// It is greedy in the sense that it goes through the node set only once
@@ -347,7 +367,7 @@ impl<Node: Idx, D: DirectedGraph<Node>> InDegreePartitionOp<Node> for D {
     /// assert_eq!(partition[0], 0..1);
     /// assert_eq!(partition[1], 1..4);
     /// ```
-    fn in_degree_partition(&self, concurrency: usize) -> Vec<Range<Node>> {
+    fn in_degree_partition(&self, concurrency: usize) -> Vec<Range<NI>> {
         let batch_size = (self.edge_count().index() as f64 / concurrency as f64).ceil();
         greedy_node_map_partition(
             |node| self.in_degree(node).index(),
@@ -361,15 +381,15 @@ impl<Node: Idx, D: DirectedGraph<Node>> InDegreePartitionOp<Node> for D {
 // Split input slice into a vector of partition.len() disjoint slices such that
 // the slice at index i in the output vector has the same length as the range at
 // index i in the input partition.
-fn split_by_partition<'a, Node: Idx, T>(
-    partition: &[Range<Node>],
+fn split_by_partition<'a, NI: Idx, T>(
+    partition: &[Range<NI>],
     slice: &'a mut [T],
 ) -> Vec<&'a mut [T]> {
     debug_assert_eq!(
         partition
             .iter()
             .map(|r| r.end - r.start)
-            .sum::<Node>()
+            .sum::<NI>()
             .index(),
         slice.len()
     );
@@ -377,7 +397,7 @@ fn split_by_partition<'a, Node: Idx, T>(
     let mut splits = Vec::with_capacity(partition.len());
 
     let mut remainder = slice;
-    let mut current_start = Node::zero();
+    let mut current_start = NI::zero();
     for range in partition.iter() {
         let next_end = range.end - current_start;
         current_start += next_end;
@@ -395,29 +415,29 @@ fn split_by_partition<'a, Node: Idx, T>(
 // that the sums of node_map(node) for each range are roughly equal. It does so
 // greedily and therefore does not guarantee an optimally balanced range-based
 // partition.
-fn greedy_node_map_partition<Node, F>(
+fn greedy_node_map_partition<NI, F>(
     node_map: F,
-    node_count: Node,
+    node_count: NI,
     batch_size: usize,
     max_batches: usize,
-) -> Vec<Range<Node>>
+) -> Vec<Range<NI>>
 where
-    F: Fn(Node) -> usize,
-    Node: Idx,
+    F: Fn(NI) -> usize,
+    NI: Idx,
 {
     let mut partitions = Vec::with_capacity(max_batches);
 
     let mut partition_size = 0;
-    let mut partition_start = Node::zero();
-    let upper_bound = node_count - Node::new(1);
+    let mut partition_start = NI::zero();
+    let upper_bound = node_count - NI::new(1);
 
-    for node in Node::zero()..node_count {
+    for node in NI::zero()..node_count {
         partition_size += node_map(node);
 
         if (partitions.len() < max_batches - 1 && partition_size >= batch_size)
             || node == upper_bound
         {
-            let partition_end = node + Node::new(1);
+            let partition_end = node + NI::new(1);
             partitions.push(partition_start..partition_end);
             partition_size = 0;
             partition_start = partition_end;
@@ -427,10 +447,15 @@ where
     partitions
 }
 
-fn relabel_by_degree<Node, G>(graph: &G) -> G
+fn relabel_by_degree<NI, G, EV>(graph: &G) -> G
 where
-    Node: Idx,
-    G: From<Csr<Node>> + UndirectedGraph<Node> + Sync,
+    NI: Idx,
+    G: From<Csr<NI, NI, EV>>
+        + Graph<NI>
+        + UndirectedDegrees<NI>
+        + UndirectedNeighborsWithValues<NI, EV>
+        + Sync,
+    EV: Copy + Ord + Sync,
 {
     let start = Instant::now();
     let degree_node_pairs = sort_by_degree_desc(graph);
@@ -453,17 +478,17 @@ where
 
 // Extracts (degree, node_id) pairs from the given graph and sorts them by
 // degree descending.
-fn sort_by_degree_desc<Node, G>(graph: &G) -> Vec<(Node, Node)>
+fn sort_by_degree_desc<NI, EV, G>(graph: &G) -> Vec<(NI, NI)>
 where
-    Node: Idx,
-    G: From<Csr<Node>> + UndirectedGraph<Node> + Sync,
+    NI: Idx,
+    G: Graph<NI> + UndirectedDegrees<NI> + UndirectedNeighborsWithValues<NI, EV> + Sync,
 {
     let node_count = graph.node_count().index();
     let mut degree_node_pairs = Vec::with_capacity(node_count);
 
     (0..node_count)
         .into_par_iter()
-        .map(Node::new)
+        .map(NI::new)
         .map(|node_id| (graph.degree(node_id), node_id))
         .collect_into_vec(&mut degree_node_pairs);
     degree_node_pairs.par_sort_unstable_by(|left, right| left.cmp(right).reverse());
@@ -475,12 +500,10 @@ where
 //
 // `degrees` maps a new node id to its degree.
 // `nodes` maps the previous node id to the new node id.
-fn unzip_degrees_and_nodes<Node: Idx>(
-    degree_node_pairs: Vec<(Node, Node)>,
-) -> (Vec<Node>, Vec<Node>) {
+fn unzip_degrees_and_nodes<NI: Idx>(degree_node_pairs: Vec<(NI, NI)>) -> (Vec<NI>, Vec<NI>) {
     let node_count = degree_node_pairs.len();
-    let mut degrees = Vec::<Node>::with_capacity(node_count);
-    let mut nodes = Vec::<Node>::with_capacity(node_count);
+    let mut degrees = Vec::<NI>::with_capacity(node_count);
+    let mut nodes = Vec::<NI>::with_capacity(node_count);
     let nodes_ptr = SharedMut::new(nodes.as_mut_ptr());
 
     (0..node_count)
@@ -492,7 +515,7 @@ fn unzip_degrees_and_nodes<Node: Idx>(
             // created from 0..node_count -- the values are all distinct and we
             // will not write into the same location in parallel
             unsafe {
-                nodes_ptr.add(node.index()).write(Node::new(n));
+                nodes_ptr.add(node.index()).write(NI::new(n));
             }
 
             degree
@@ -508,45 +531,42 @@ fn unzip_degrees_and_nodes<Node: Idx>(
 }
 
 // Relabel target ids according to the given node mapping and offsets.
-fn relabel_targets<Node, G>(graph: &G, nodes: Vec<Node>, offsets: &[Node]) -> Vec<Node>
+fn relabel_targets<NI, EV, G>(graph: &G, nodes: Vec<NI>, offsets: &[NI]) -> Vec<Target<NI, EV>>
 where
-    Node: Idx,
-    G: From<Csr<Node>> + UndirectedGraph<Node> + Sync,
+    NI: Idx,
+    G: Graph<NI> + UndirectedNeighborsWithValues<NI, EV> + Sync,
+    EV: Copy + Ord + Sync,
 {
     let node_count = graph.node_count().index();
     let edge_count = offsets[node_count].index();
-    let mut targets = Vec::<Node>::with_capacity(edge_count);
+    let mut targets = Vec::<Target<NI, EV>>::with_capacity(edge_count);
     let targets_ptr = SharedMut::new(targets.as_mut_ptr());
 
-    (0..node_count)
-        .into_par_iter()
-        .map(Node::new)
-        .for_each(|u| {
-            let new_u = nodes[u.index()];
-            let start_offset = offsets[new_u.index()].index();
-            let mut end_offset = start_offset;
+    (0..node_count).into_par_iter().map(NI::new).for_each(|u| {
+        let new_u = nodes[u.index()];
+        let start_offset = offsets[new_u.index()].index();
+        let mut end_offset = start_offset;
 
-            for &v in graph.neighbors(u) {
-                let new_v = nodes[v.index()];
-                // SAFETY: a node u is processed by at most one thread. We write
-                // into a non-overlapping range defined by the offsets for that
-                // node. No two threads will write into the same range.
-                unsafe {
-                    targets_ptr.add(end_offset).write(new_v);
-                }
-                end_offset += 1;
-            }
-
-            // SAFETY: start_offset..end_offset is a non-overlapping range for
-            // a node u which is processed by exactly one thread.
+        for &v in graph.neighbors_with_values(u) {
+            let new_v = nodes[v.target.index()];
+            // SAFETY: a node u is processed by at most one thread. We write
+            // into a non-overlapping range defined by the offsets for that
+            // node. No two threads will write into the same range.
             unsafe {
-                std::slice::from_raw_parts_mut(
-                    targets_ptr.add(start_offset),
-                    end_offset - start_offset,
-                )
+                targets_ptr
+                    .add(end_offset)
+                    .write(Target::new(new_v, v.value));
             }
-            .sort_unstable();
-        });
+            end_offset += 1;
+        }
+
+        // SAFETY: start_offset..end_offset is a non-overlapping range for
+        // a node u which is processed by exactly one thread.
+        unsafe {
+            std::slice::from_raw_parts_mut(targets_ptr.add(start_offset), end_offset - start_offset)
+        }
+        .sort_unstable();
+    });
 
     // SAFETY: we inserted every relabeled target id of which there are edge_count many.
     unsafe {
@@ -560,10 +580,15 @@ where
 mod tests {
     use crate::{
         builder::GraphBuilder, graph::csr::UndirectedCsrGraph, graph_ops::unzip_degrees_and_nodes,
+        UndirectedNeighbors,
     };
 
     use super::*;
 
+    fn t<T>(t: T) -> Target<T, ()> {
+        Target::new(t, ())
+    }
+
     #[test]
     fn split_by_partition_3_parts() {
         let partition = vec![0..2, 2..5, 5..10];
