@@ -2,6 +2,7 @@ pub use crate::builder::GraphBuilder;
 
 pub use crate::graph::csr::CsrLayout;
 pub use crate::graph::csr::DirectedCsrGraph;
+pub use crate::graph::csr::Target;
 pub use crate::graph::csr::UndirectedCsrGraph;
 
 pub use crate::graph::labeled::DirectedNodeLabeledCsrGraph;
@@ -19,9 +20,13 @@ pub use crate::index::Idx;
 
 pub use crate::input::*;
 
-pub use crate::DirectedGraph;
+pub use crate::DirectedDegrees;
+pub use crate::DirectedNeighbors;
+pub use crate::DirectedNeighborsWithValues;
 pub use crate::Graph;
 pub use crate::NodeLabeledGraph;
-pub use crate::UndirectedGraph;
+pub use crate::UndirectedDegrees;
+pub use crate::UndirectedNeighbors;
+pub use crate::UndirectedNeighborsWithValues;
 
 pub use crate::Error;
